@@ -2,6 +2,7 @@ import warnings
 import pathlib as path
 import shutil
 import cdflib
+import datetime
 
 import numpy as np
 import pandas as pd
@@ -228,9 +229,20 @@ def cdf2xr(cdf, index_key, dtimeindex=True, badvalues=None, ignore=None,
     out : :class:`xarray.DataArray` or `xarray.Dataset`
         xarray object containing data from the open CDF file.
     """
-    vars, keys, index = generate_vars_keys_index(
-        cdf, index_key, dtimeindex=True, badvalues=None, ignore=None,
-        include=None)
+    if badvalues is not None:
+        warnings.warn('The badvalues argument is decprecated, as bad values '
+                      'are now automatically recognised using the FILLVAL CDF '
+                      'attribute.', DeprecationWarning)
+    if include is not None:
+        if ignore is not None:
+            raise ValueError('ignore and include are incompatible keywords')
+        if isinstance(include, str):
+            include = [include]
+        if index_key not in include:
+            include.append(index_key)
+
+    index = get_index(cdf, index_key)
+
     ind = np.intersect1d(index, index, return_indices=True)[1]
 
     # If none of the required data in current CDF file, move on to the next one
@@ -241,6 +253,36 @@ def cdf2xr(cdf, index_key, dtimeindex=True, badvalues=None, ignore=None,
     if not product_list:
         data = xr.Dataset({})
 
+        npoints = index.shape[0]
+
+        var_list = _get_cdf_vars(cdf)
+        keys = {}
+        # Get mapping from each attr to sub-variables
+        for cdf_key in var_list:
+            if ignore:
+                if cdf_key in ignore:
+                    continue
+            elif include:
+                if cdf_key not in include:
+                    continue
+            if cdf_key == 'Epoch':
+                keys[cdf_key] = 'Time'
+            else:
+                keys[cdf_key] = cdf_key
+        # Remove index key, as we have already used it to create the index
+        keys.pop(index_key)
+        # Remove keys for data that doesn't have the right shape to load in CDF
+        # Mapping of keys to variable data
+        vars = {cdf_key: cdf.varget(cdf_key) for cdf_key in keys.copy()}
+        for cdf_key in keys:
+            var = vars[cdf_key]
+            if type(var) is np.ndarray:
+                key_shape = var.shape
+                if len(key_shape) == 0 or key_shape[0] != npoints:
+                    vars.pop(cdf_key)
+            else:
+                vars.pop(cdf_key)
+
         # Loop through each key and put data into the dataset
         for cdf_key in vars:
             df_key = keys[cdf_key]
@@ -482,32 +524,8 @@ def generate_vars_keys_index(cdf, index_key, dtimeindex=True, badvalues=None,
             include.append(index_key)
 
     # Extract index values
-    index = cdf.varget(index_key)
-    try:
-        # If there are multiple indexes, take the first one
-        # TODO: this is just plain wrong, there should be a way to get all
-        # the indexes out
-        index = index[...][:, 0]
-    except IndexError:
-        pass
+    index = get_index(cdf, index_key)
 
-    if dtimeindex:
-        index = cdflib.epochs.CDFepoch.breakdown(index, to_np=True)
-        index_df = pd.DataFrame({'year': index[:, 0],
-                                 'month': index[:, 1],
-                                 'day': index[:, 2],
-                                 'hour': index[:, 3],
-                                 'minute': index[:, 4],
-                                 'second': index[:, 5],
-                                 'ms': index[:, 6],
-                                 })
-        # Not all CDFs store pass milliseconds
-        try:
-            index_df['us'] = index[:, 7]
-            index_df['ns'] = index[:, 8]
-        except IndexError:
-            pass
-        index = pd.DatetimeIndex(pd.to_datetime(index_df), name='Time')
     npoints = index.shape[0]
 
     var_list = _get_cdf_vars(cdf)
@@ -564,3 +582,75 @@ def convert_units_to_str(dict_input):
             raise ValueError('Problem to convert the units. There are not'
                              'astropy units or quantity')
     return dict_out
+
+
+def get_index(cdf, index_key, dtimeindex=True):
+    # Extract index values
+    try:
+        index_ = cdf.varget(index_key)[...][:, 0]
+    except IndexError:
+        index_ = cdf.varget(index_key)[...]
+    except ValueError:
+        keys = get_cdf_keys(cdf)
+        index_key = [x for x in list(keys) if index_key.casefold()
+                     in x.casefold()][0]
+        try:
+            index_ = cdf.varget(index_key)[...][:, 0]
+        except IndexError:
+            index_ = cdf.varget(index_key)[...]
+
+        message = (f"'{index_key}' has been automatically selected as"
+                   f" the time index in the present CDF")
+        warnings.warn(message)
+    try:
+        utc_comp = cdflib.cdfepoch.breakdown(index_, to_np=True)
+        if utc_comp.shape[1] == 9:
+            millis = utc_comp[:, 6] * (10**3)
+            micros = utc_comp[:, 8] * (10**2)
+            nanos = utc_comp[:, 7]
+            utc_comp[:, 6] = millis + micros + nanos
+            utc_comp = np.delete(utc_comp, np.s_[-2:], axis=1)
+        try:
+            index = np.asarray([datetime.datetime(*x) for x in utc_comp])
+        except ValueError:
+            utc_comp[:, 6] -= micros
+            index = np.asarray([datetime.datetime(*x) for x in utc_comp])
+    except Exception:
+        index = index_
+    '''
+    if dtimeindex:
+        index = cdflib.epochs.CDFepoch.breakdown(index, to_np=True)
+        index_df = pd.DataFrame({'year': index[:, 0],
+                                 'month': index[:, 1],
+                                 'day': index[:, 2],
+                                 'hour': index[:, 3],
+                                 'minute': index[:, 4],
+                                 'second': index[:, 5],
+                                 'ms': index[:, 6],
+                                 })
+        # Not all CDFs store pass milliseconds
+        try:
+            index_df['us'] = index[:, 7]
+            index_df['ns'] = index[:, 8]
+        except IndexError:
+            pass
+        index = pd.DatetimeIndex(pd.to_datetime(index_df), name='Time')
+    '''
+
+    return index
+
+
+def get_cdf_keys(cdf):
+    var_list = []
+    for attr in list(cdf.cdf_info().keys()):
+        if 'variable' in attr.lower():
+            if len(cdf.cdf_info()[attr]) > 0:
+                var_list += [attr]
+
+    keys = {}
+    for attr in var_list:
+        for cdf_key in cdf.cdf_info()[attr]:
+            keys[cdf_key] = cdf_key
+
+    return keys
+
